           (*-------------------------------------------*
            |                CSP-Prover                 |
            |              December 2004                |
            |              February 2005  (modified)    |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_rule_rw = CSP_law_etc:

(*****************************************************************

         1.
         2.
         3.
         4.

 *****************************************************************)

(*-------------------------------------------------------* 
 |                                                       |
 |       use these rules "csp_rule_..." as follows:      |
 |                                                       |
 |     lemmas new_rule = rule[THEN csp_rule_... ] or     |
 |     lemmas new_rule = csp_rule_...[OF rule]           |
 |                                                       |
 *-------------------------------------------------------*)

lemma csp_rule_rw_left_eq:
  "R1 =F R2 ==> (R1 =F R0) = (R2 =F R0)"
by (simp add: eqF_def)

lemma csp_rule_rw_left_ref:
  "R1 =F R2 ==> (R1 <=F R0) = (R2 <=F R0)"
by (simp add: eqF_def refF_def)

lemma csp_rule_rw_right_eq:
  "R1 =F R2 ==> (R0 =F R1) = (R0 =F R2)"
by (simp add: eqF_def)

lemma csp_rule_rw_right_ref:
  "R1 =F R2 ==> (R0 <=F R1) = (R0 <=F R2)"
by (simp add: eqF_def refF_def)

lemmas csp_rule_rw_left  = csp_rule_rw_left_eq csp_rule_rw_left_ref
lemmas csp_rule_rw_right = csp_rule_rw_right_eq csp_rule_rw_right_ref
lemmas csp_rule_rw       = csp_rule_rw_left csp_rule_rw_right

lemmas csp_rule_rev_rw   = csp_rule_rw[THEN sym]


(*********************************************************
                    Step laws
 *********************************************************)

lemmas csp_step = csp_STOP_step       csp_Act_prefix_step 
                  csp_Ext_choice_step csp_Parallel_step
                  csp_Hiding_step     csp_Renaming_step
                  csp_Seq_compo_step

lemmas csp_step_rw_left_eq   = csp_step[THEN csp_rule_rw_left_eq]
lemmas csp_step_rw_left_ref  = csp_step[THEN csp_rule_rw_left_ref]
lemmas csp_step_rw_right_eq  = csp_step[THEN csp_rule_rw_right_eq]
lemmas csp_step_rw_right_ref = csp_step[THEN csp_rule_rw_right_ref]

lemmas csp_step_rw_left      = csp_step_rw_left_eq csp_step_rw_left_ref
lemmas csp_step_rw_right     = csp_step_rw_right_eq csp_step_rw_right_ref

lemmas csp_step_rw           = csp_step_rw_left csp_step_rw_right

(*********************************************************
          light Step laws for Act_prefix and STOP
 *********************************************************)

lemmas csp_light_step = csp_STOP_step  csp_Act_prefix_step 

lemmas csp_light_step_rw_left_eq   = csp_light_step[THEN csp_rule_rw_left_eq]
lemmas csp_light_step_rw_left_ref  = csp_light_step[THEN csp_rule_rw_left_ref]
lemmas csp_light_step_rw_right_eq  = csp_light_step[THEN csp_rule_rw_right_eq]
lemmas csp_light_step_rw_right_ref = csp_light_step[THEN csp_rule_rw_right_ref]

lemmas csp_light_step_rw_left      = csp_light_step_rw_left_eq csp_light_step_rw_left_ref
lemmas csp_light_step_rw_right     = csp_light_step_rw_right_eq csp_light_step_rw_right_ref

lemmas csp_light_step_rw           = csp_light_step_rw_left csp_light_step_rw_right

(*********************************************************
                       SKIP laws
 *********************************************************)

lemmas csp_SKIP = csp_Seq_compo_unit_l csp_Seq_compo_unit_r
                  csp_Parallel_preterm_l csp_Parallel_preterm_r
                  csp_Parallel_term csp_SKIP_Hiding_Id
                  csp_SKIP_Renaming_Id

lemmas csp_SKIP_rw_left_eq   = csp_SKIP[THEN csp_rule_rw_left_eq]
lemmas csp_SKIP_rw_left_ref  = csp_SKIP[THEN csp_rule_rw_left_ref]
lemmas csp_SKIP_rw_right_eq  = csp_SKIP[THEN csp_rule_rw_right_eq]
lemmas csp_SKIP_rw_right_ref = csp_SKIP[THEN csp_rule_rw_right_ref]

lemmas csp_SKIP_rw_left      = csp_SKIP_rw_left_eq csp_SKIP_rw_left_ref
lemmas csp_SKIP_rw_right     = csp_SKIP_rw_right_eq csp_SKIP_rw_right_ref

lemmas csp_SKIP_rw           = csp_SKIP_rw_left csp_SKIP_rw_right

(*********************************************************
                 name unwinding function
 *********************************************************)

lemmas csp_unwind_rw_left_eq   = csp_unwind[THEN csp_rule_rw_left_eq]
lemmas csp_unwind_rw_left_ref  = csp_unwind[THEN csp_rule_rw_left_ref]
lemmas csp_unwind_rw_right_eq  = csp_unwind[THEN csp_rule_rw_right_eq]
lemmas csp_unwind_rw_right_ref = csp_unwind[THEN csp_rule_rw_right_ref]

lemmas csp_unwind_rw_left      = csp_unwind_rw_left_eq csp_unwind_rw_left_ref
lemmas csp_unwind_rw_right     = csp_unwind_rw_right_eq csp_unwind_rw_right_ref

lemmas csp_unwind_rw           = csp_unwind_rw_left csp_unwind_rw_right

(*** cpo ***)

lemmas csp_unwind_cpo_rw_left_eq   = csp_unwind_cpo[THEN csp_rule_rw_left_eq]
lemmas csp_unwind_cpo_rw_left_ref  = csp_unwind_cpo[THEN csp_rule_rw_left_ref]
lemmas csp_unwind_cpo_rw_right_eq  = csp_unwind_cpo[THEN csp_rule_rw_right_eq]
lemmas csp_unwind_cpo_rw_right_ref = csp_unwind_cpo[THEN csp_rule_rw_right_ref]

lemmas csp_unwind_cpo_rw_left      = csp_unwind_cpo_rw_left_eq csp_unwind_cpo_rw_left_ref
lemmas csp_unwind_cpo_rw_right     = csp_unwind_cpo_rw_right_eq csp_unwind_cpo_rw_right_ref

lemmas csp_unwind_cpo_rw           = csp_unwind_cpo_rw_left csp_unwind_cpo_rw_right

(*********************************************************
                     dist laws
 *********************************************************)

lemmas csp_dist = csp_Ext_choice_dist_l csp_Ext_choice_dist_r
                  csp_Parallel_dist_l   csp_Parallel_dist_r
                  csp_Hiding_dist       csp_Renaming_dist
                  csp_Seq_compo_dist

lemmas csp_dist_rw_left_eq   = csp_dist[THEN csp_rule_rw_left_eq]
lemmas csp_dist_rw_right_eq  = csp_dist[THEN csp_rule_rw_right_eq]
lemmas csp_dist_rw_left_ref  = csp_dist[THEN csp_rule_rw_left_ref]
lemmas csp_dist_rw_right_ref = csp_dist[THEN csp_rule_rw_right_ref]

lemmas csp_dist_rw_left      = csp_dist_rw_left_eq csp_dist_rw_left_ref
lemmas csp_dist_rw_right     = csp_dist_rw_right_eq csp_dist_rw_right_ref

lemmas csp_dist_rw           = csp_dist_rw_left csp_dist_rw_right

(*********************************************************
                     Rep_dist laws
 *********************************************************)

lemmas csp_Rep_dist = csp_Ext_choice_Rep_dist_l csp_Ext_choice_Rep_dist_r
                  csp_Parallel_Rep_dist_l   csp_Parallel_Rep_dist_r
                  csp_Hiding_Rep_dist       csp_Renaming_Rep_dist
                  csp_Seq_compo_Rep_dist

lemmas csp_Rep_dist_rw_left_eq   = csp_Rep_dist[THEN csp_rule_rw_left_eq]
lemmas csp_Rep_dist_rw_right_eq  = csp_Rep_dist[THEN csp_rule_rw_right_eq]
lemmas csp_Rep_dist_rw_left_ref  = csp_Rep_dist[THEN csp_rule_rw_left_ref]
lemmas csp_Rep_dist_rw_right_ref = csp_Rep_dist[THEN csp_rule_rw_right_ref]

lemmas csp_Rep_dist_rw_left      = csp_Rep_dist_rw_left_eq csp_Rep_dist_rw_left_ref
lemmas csp_Rep_dist_rw_right     = csp_Rep_dist_rw_right_eq csp_Rep_dist_rw_right_ref

lemmas csp_Rep_dist_rw           = csp_Rep_dist_rw_left csp_Rep_dist_rw_right

(*********************************************************
                    Conditional
 *********************************************************)

lemmas csp_Conditional_split_rw_left_eq = 
       csp_Conditional_split[THEN csp_rule_rw_left_eq]
lemmas csp_Conditional_split_rw_left_ref =
       csp_Conditional_split[THEN csp_rule_rw_left_ref]
lemmas csp_Conditional_split_rw_right_eq =
       csp_Conditional_split[THEN csp_rule_rw_right_eq]
lemmas csp_Conditional_split_rw_right_ref =
       csp_Conditional_split[THEN csp_rule_rw_right_ref]

lemmas csp_Conditional_split_rw_left  = csp_Conditional_split_rw_left_eq 
                                        csp_Conditional_split_rw_left_ref
lemmas csp_Conditional_split_rw_right = csp_Conditional_split_rw_right_eq
                                         csp_Conditional_split_rw_right_ref

lemmas csp_Conditional_split_rw       = csp_Conditional_split_rw_left
                                        csp_Conditional_split_rw_right

(*** True or False ***)

lemmas csp_Conditional = csp_Conditional_True csp_Conditional_False

lemmas csp_Conditional_rw_left_eq   = csp_Conditional[THEN csp_rule_rw_left_eq]
lemmas csp_Conditional_rw_left_ref  = csp_Conditional[THEN csp_rule_rw_left_ref]
lemmas csp_Conditional_rw_right_eq  = csp_Conditional[THEN csp_rule_rw_right_eq]
lemmas csp_Conditional_rw_right_ref = csp_Conditional[THEN csp_rule_rw_right_ref]

lemmas csp_Conditional_rw_left      = csp_Conditional_rw_left_eq 
                                      csp_Conditional_rw_left_ref
lemmas csp_Conditional_rw_right     = csp_Conditional_rw_right_eq 
                                      csp_Conditional_rw_right_ref

lemmas csp_Conditional_rw           = csp_Conditional_rw_left 
                                      csp_Conditional_rw_right

(*********************************************************
                      Int_simp
 *********************************************************)

lemmas csp_Int_simp = csp_DIV_unwind 
                      csp_Rep_int_choice_unique 
                      csp_Int_choice_idem_weak

lemmas csp_Int_simp_rw_left_eq   = csp_Int_simp[THEN csp_rule_rw_left_eq]
lemmas csp_Int_simp_rw_left_ref  = csp_Int_simp[THEN csp_rule_rw_left_ref]
lemmas csp_Int_simp_rw_right_eq  = csp_Int_simp[THEN csp_rule_rw_right_eq]
lemmas csp_Int_simp_rw_right_ref = csp_Int_simp[THEN csp_rule_rw_right_ref]

lemmas csp_Int_simp_rw_left      = csp_Int_simp_rw_left_eq csp_Int_simp_rw_left_ref
lemmas csp_Int_simp_rw_right     = csp_Int_simp_rw_right_eq csp_Int_simp_rw_right_ref

lemmas csp_Int_simp_rw           = csp_Int_simp_rw_left csp_Int_simp_rw_right

(*********************************************************
                      Ext_simp
 *********************************************************)

lemmas csp_Ext_simp = csp_Ext_choice_unit_l_weak
                      csp_Ext_choice_unit_r_weak

lemmas csp_Ext_simp_rw_left_eq   = csp_Ext_simp[THEN csp_rule_rw_left_eq]
lemmas csp_Ext_simp_rw_left_ref  = csp_Ext_simp[THEN csp_rule_rw_left_ref]
lemmas csp_Ext_simp_rw_right_eq  = csp_Ext_simp[THEN csp_rule_rw_right_eq]
lemmas csp_Ext_simp_rw_right_ref = csp_Ext_simp[THEN csp_rule_rw_right_ref]

lemmas csp_Ext_simp_rw_left      = csp_Ext_simp_rw_left_eq csp_Ext_simp_rw_left_ref
lemmas csp_Ext_simp_rw_right     = csp_Ext_simp_rw_right_eq csp_Ext_simp_rw_right_ref

lemmas csp_Ext_simp_rw           = csp_Ext_simp_rw_left csp_Ext_simp_rw_right

(*********************************************************
                    Timeout_simp
 *********************************************************)

lemmas csp_Mix_simp = csp_STOP_Timeout_weak

lemmas csp_Mix_simp_rw_left_eq   = csp_Mix_simp[THEN csp_rule_rw_left_eq]
lemmas csp_Mix_simp_rw_left_ref  = csp_Mix_simp[THEN csp_rule_rw_left_ref]
lemmas csp_Mix_simp_rw_right_eq  = csp_Mix_simp[THEN csp_rule_rw_right_eq]
lemmas csp_Mix_simp_rw_right_ref = csp_Mix_simp[THEN csp_rule_rw_right_ref]

lemmas csp_Mix_simp_rw_left      = csp_Mix_simp_rw_left_eq csp_Mix_simp_rw_left_ref
lemmas csp_Mix_simp_rw_right     = csp_Mix_simp_rw_right_eq csp_Mix_simp_rw_right_ref

lemmas csp_Mix_simp_rw           = csp_Mix_simp_rw_left csp_Mix_simp_rw_right

(*********************************************************
                      csp_simp
 *********************************************************)

lemmas csp_simp  = csp_Int_simp csp_Ext_simp csp_Mix_simp

lemmas csp_simp_rw_left  = csp_Int_simp_rw_left csp_Ext_simp_rw_left
                           csp_Mix_simp_rw_left
lemmas csp_simp_rw_right = csp_Int_simp_rw_right csp_Ext_simp_rw_right
                           csp_Mix_simp_rw_right
lemmas csp_simp_rw       = csp_simp_rw_left csp_simp_rw_right

(*-------------------------------------------------------* 
 |                                                       |
 |      decompostion controlled by Not_Decompo_Flag      |
 |                                                       |
 *-------------------------------------------------------*)

consts Not_Decompo_Flag :: bool
defs   Not_Decompo_Flag_def : "Not_Decompo_Flag == True"

lemma on_Not_Decompo_Flag: "Not_Decompo_Flag & R ==> R"
by (simp)

lemma off_Not_Decompo_Flag: "R ==> Not_Decompo_Flag & R"
by (simp add: Not_Decompo_Flag_def)

lemma off_Not_Decompo_Flag_True: "Not_Decompo_Flag"
by (simp add: Not_Decompo_Flag_def)

(*----------------------------------------------*
 |   "Not_Decompo_Flag" is used for avoiding    |
 |    infinite decompositions in csp_tactics.   |
 *----------------------------------------------*)

lemma csp_rule_flag_left_eq:
  "(R1 =F R2) ==> (Not_Decompo_Flag & (R2 =F R0) ==> (R1 =F R0))"
by (simp add: eqF_def Not_Decompo_Flag_def)

lemma csp_rule_flag_left_ref:
  "R1 <=F R2 ==> (Not_Decompo_Flag & (R2 <=F R0) ==> (R1 <=F R0))"
by (simp add: eqF_def refF_def Not_Decompo_Flag_def)

lemma csp_rule_flag_right_eq:
  "R2 =F R1 ==> (Not_Decompo_Flag & (R0 =F R2) ==> (R0 =F R1))"
by (simp add: eqF_def Not_Decompo_Flag_def)

lemma csp_rule_flag_right_ref:
  "R2 <=F R1 ==> (Not_Decompo_Flag & (R0 <=F R2) ==> (R0 <=F R1))"
by (simp add: eqF_def refF_def Not_Decompo_Flag_def)

lemmas csp_rule_flag_left  = csp_rule_flag_left_eq csp_rule_flag_left_ref
lemmas csp_rule_flag_right = csp_rule_flag_right_eq csp_rule_flag_right_ref
lemmas csp_rule_flag       = csp_rule_flag_left csp_rule_flag_right

(*** no flag ***)

lemma csp_rule_tr_left_eq:
  "R1 =F R2 ==> ((R2 =F R0) ==> (R1 =F R0))"
by (simp add: eqF_def)

lemma csp_rule_tr_left_ref:
  "R1 <=F R2 ==> ((R2 <=F R0) ==> (R1 <=F R0))"
by (simp add: eqF_def refF_def)

lemma csp_rule_tr_right_eq:
  "R2 =F R1 ==> ((R0 =F R2) ==> (R0 =F R1))"
by (simp add: eqF_def)

lemma csp_rule_tr_right_ref:
  "R2 <=F R1 ==> ((R0 <=F R2) ==> (R0 <=F R1))"
by (simp add: eqF_def refF_def)

lemmas csp_rule_tr_left  = csp_rule_tr_left_eq csp_rule_tr_left_ref
lemmas csp_rule_tr_right = csp_rule_tr_right_eq csp_rule_tr_right_ref
lemmas csp_rule_tr       = csp_rule_tr_left csp_rule_tr_right

lemmas csp_rule_tr_eq   = csp_rule_tr_left_eq csp_rule_tr_right_eq
lemmas csp_rule_tr_ref  = csp_rule_tr_left_ref csp_rule_tr_right_ref

(*------------------------------------------------*
 |           laws for monotonicity (all)          |
 *------------------------------------------------*)

lemmas csp_free_mono_left  = csp_free_mono[THEN csp_rule_tr_left_ref]
lemmas csp_free_mono_right = csp_free_mono[THEN csp_rule_tr_right_ref]

lemmas csp_mono_left  = csp_mono[THEN csp_rule_tr_left_ref]
lemmas csp_mono_right = csp_mono[THEN csp_rule_tr_right_ref]

lemmas csp_free_mono_flag_left  = csp_free_mono[THEN csp_rule_flag_left_ref]
lemmas csp_free_mono_flag_right = csp_free_mono[THEN csp_rule_flag_right_ref]

lemmas csp_mono_flag_left  = csp_mono[THEN csp_rule_flag_left_ref]
lemmas csp_mono_flag_right = csp_mono[THEN csp_rule_flag_right_ref]

(*------------------------------------------------*
 |            laws for congruence (all)           |
 *------------------------------------------------*)

lemmas csp_free_cong_left  = csp_free_cong[THEN csp_rule_tr_left_eq]
lemmas csp_free_cong_right = csp_free_cong[THEN csp_rule_tr_right_eq]

lemmas csp_cong_left  = csp_cong[THEN csp_rule_tr_left_eq]
lemmas csp_cong_right = csp_cong[THEN csp_rule_tr_right_eq]

lemmas csp_free_cong_flag_left  = csp_free_cong[THEN csp_rule_flag_left_eq]
lemmas csp_free_cong_flag_right = csp_free_cong[THEN csp_rule_flag_right_eq]

lemmas csp_cong_flag_left  = csp_cong[THEN csp_rule_flag_left_eq]
lemmas csp_cong_flag_right = csp_cong[THEN csp_rule_flag_right_eq]

(*------------------------------------------------*
 |           laws for decompo (ALL)               |
 *------------------------------------------------*)

lemmas csp_free_decompo_left  = csp_free_mono_left csp_free_cong_left
lemmas csp_free_decompo_right = csp_free_mono_right csp_free_cong_right

lemmas csp_decompo_left  = csp_mono_left csp_cong_left
lemmas csp_decompo_right = csp_mono_right csp_cong_right

lemmas csp_free_decompo_flag_left  = csp_free_mono_flag_left csp_free_cong_flag_left
lemmas csp_free_decompo_flag_right = csp_free_mono_flag_right csp_free_cong_flag_right

lemmas csp_decompo_flag_left  = csp_mono_flag_left csp_cong_flag_left
lemmas csp_decompo_flag_right = csp_mono_flag_right csp_cong_flag_right

(*------------------------------------------------*
 |              trans with Flag                   |
 *------------------------------------------------*)

(*** rewrite (eq) ***)

lemma csp_rw_flag_left_eq:
  "[| R1 =F R2 ; Not_Decompo_Flag & R2 =F R3 |] ==> R1 =F R3"
by (simp add: eqF_def Not_Decompo_Flag_def)

lemma csp_rw_flag_left_ref:
  "[| R1 =F R2 ; Not_Decompo_Flag & R2 <=F R3 |] ==> R1 <=F R3"
by (simp add: refF_def eqF_def Not_Decompo_Flag_def)

lemmas csp_rw_flag_left = csp_rw_flag_left_eq csp_rw_flag_left_ref

lemma csp_rw_flag_right_eq:
  "[| R3 =F R2 ; Not_Decompo_Flag & R1 =F R2 |] ==> R1 =F R3"
by (simp add: eqF_def Not_Decompo_Flag_def)

lemma csp_rw_flag_right_ref:
  "[| R3 =F R2 ; Not_Decompo_Flag & R1 <=F R2 |] ==> R1 <=F R3"
by (simp add: refF_def eqF_def Not_Decompo_Flag_def)

lemmas csp_rw_flag_right = csp_rw_flag_right_eq csp_rw_flag_right_ref

(*** rewrite (ref) ***)

lemma csp_tr_flag_left_eq: "[| R1 =F R2 ; Not_Decompo_Flag & R2 =F R3 |] ==> R1 =F R3"
by (simp add: eqF_def Not_Decompo_Flag_def)

lemma csp_tr_flag_left_ref: "[| R1 <=F R2 ; Not_Decompo_Flag & R2 <=F R3 |] ==> R1 <=F R3"
by (simp add: refF_def eqF_def Not_Decompo_Flag_def)

lemmas csp_tr_flag_left = csp_tr_flag_left_eq csp_tr_flag_left_ref

lemma csp_tr_flag_right_eq: "[| R2 =F R3 ; Not_Decompo_Flag & R1 =F R2 |] ==> R1 =F R3"
by (simp add: eqF_def Not_Decompo_Flag_def)

lemma csp_tr_flag_right_ref: "[| R2 <=F R3 ; Not_Decompo_Flag & R1 <=F R2 |] ==> R1 <=F R3"
by (simp add: refF_def eqF_def Not_Decompo_Flag_def)

lemmas csp_tr_flag_right = csp_tr_flag_right_eq csp_tr_flag_right_ref

end
